@@ -173,10 +173,12 @@ class TestParse < Test::Unit::TestCase
     end
 
     c = Class.new
-    assert_raise(SyntaxError) do
+    assert_nothing_raised(SyntaxError) do
       eval <<-END, nil, __FILE__, __LINE__+1
+      if false
         c::FOO &= 1
         ::FOO &= 1
+      end
       END
     end
 
@@ -204,9 +206,9 @@ class TestParse < Test::Unit::TestCase
       END
     end
 
-    assert_raise(SyntaxError) do
+    assert_nothing_raised(SyntaxError) do
       eval <<-END, nil, __FILE__, __LINE__+1
-        class Foo Bar; end
+        class Foo 1; end
       END
     end
   end
@@ -216,7 +218,6 @@ class TestParse < Test::Unit::TestCase
     def o.>(x); x; end
     def o./(x); x; end
 
-    a = nil
     assert_nothing_raised do
       o.instance_eval <<-END, __FILE__, __LINE__+1
         undef >, /
@@ -359,10 +360,40 @@ class TestParse < Test::Unit::TestCase
     assert_equal("foo 1 bar", "foo #$1 bar")
   end
 
+  def test_dstr_disallowed_variable
+    bug8375 = '[ruby-core:54885] [Bug #8375]'
+    %w[@ @1 @. @@ @@1 @@. $ $%].each do |src|
+      src = '#'+src+' '
+      str = assert_nothing_raised(SyntaxError, "#{bug8375} #{src.dump}") do
+        break eval('"'+src+'"')
+      end
+      assert_equal(src, str, bug8375)
+    end
+  end
+
   def test_dsym
     assert_nothing_raised { eval(':""') }
   end
 
+  def assert_disallowed_variable(type, noname, *invalid)
+    assert_syntax_error(noname, "`#{noname}' without identifiers is not allowed as #{type} variable name")
+    invalid.each do |name|
+      assert_syntax_error(name, "`#{name}' is not allowed as #{type} variable name")
+    end
+  end
+
+  def test_disallowed_instance_variable
+    assert_disallowed_variable("an instance", *%w[@ @1 @.])
+  end
+
+  def test_disallowed_class_variable
+    assert_disallowed_variable("a class", *%w[@@ @@1 @@.])
+  end
+
+  def test_disallowed_gloal_variable
+    assert_disallowed_variable("a global", *%w[$ $%])
+  end
+
   def test_arg2
     o = Object.new
 
@@ -525,13 +556,14 @@ class TestParse < Test::Unit::TestCase
       )
     end
 
-    assert_raise(SyntaxError) do
-      eval %q(
+    assert_nothing_raised(SyntaxError) do
+      x = eval %q(
 <<FOO
 #$
 FOO
       )
     end
+    assert_equal "\#$\n", x
 
     assert_raise(SyntaxError) do
       eval %Q(
@@ -551,14 +583,15 @@ FOO
       )
     end
 
-    assert_raise(SyntaxError) do
-      eval %q(
+    assert_nothing_raised(SyntaxError) do
+      x = eval %q(
 <<FOO
 #$
 foo
 FOO
       )
     end
+    assert_equal "\#$\nfoo\n", x
 
     assert_nothing_raised do
       eval "x = <<""FOO\r\n1\r\nFOO"
@@ -636,15 +669,20 @@ x = __ENCODING__
 
   def test_invalid_instance_variable
     assert_raise(SyntaxError) { eval('@#') }
+    assert_raise(SyntaxError) { eval('@') }
   end
 
   def test_invalid_class_variable
     assert_raise(SyntaxError) { eval('@@1') }
+    assert_raise(SyntaxError) { eval('@@') }
   end
 
   def test_invalid_char
+    bug10117 = '[ruby-core:64243] [Bug #10117]'
+    invalid_char = /Invalid char `\\x01'/
     x = 1
-    assert_equal(1, eval("\x01x"))
+    assert_in_out_err(%W"-e \x01x", "", [], invalid_char, bug10117)
+    assert_syntax_error("\x01x", invalid_char, bug10117)
     assert_equal(nil, eval("\x04x"))
   end
 
@@ -737,13 +775,7 @@ x = __ENCODING__
       eval %q(1; next; 2)
     end
 
-    o = Object.new
-    assert_nothing_raised do
-      eval <<-END, nil, __FILE__, __LINE__+1
-        x = def o.foo; end
-      END
-    end
-    assert_equal(14, $stderr.string.lines.to_a.size)
+    assert_equal(13, $stderr.string.lines.to_a.size)
     $stderr = stderr
   end
 
@@ -807,26 +839,6 @@ x = __ENCODING__
     end
   end
 
-  def test_intern
-    assert_equal(':""', ''.intern.inspect)
-    assert_equal(':$foo', '$foo'.intern.inspect)
-    assert_equal(':"!foo"', '!foo'.intern.inspect)
-    assert_equal(':"foo=="', "foo==".intern.inspect)
-  end
-
-  def test_all_symbols
-    x = Symbol.all_symbols
-    assert_kind_of(Array, x)
-    assert(x.all? {|s| s.is_a?(Symbol) })
-  end
-
-  def test_is_class_id
-    c = Class.new
-    assert_raise(NameError) do
-      c.instance_eval { remove_class_variable(:@var) }
-    end
-  end
-
   def test_method_block_location
     bug5614 = '[ruby-core:40936]'
     expected = nil
@@ -840,4 +852,30 @@ x = __ENCODING__
     actual = e.backtrace.first[/\A#{Regexp.quote(__FILE__)}:(\d+):/o, 1].to_i
     assert_equal(expected, actual, bug5614)
   end
+
+  def test_shadowing_variable
+    assert_warning(/shadowing outer local variable/) {eval("a=1; tap {|a|}")}
+    a = "\u{3042}"
+    assert_warning(/#{a}/o) {eval("#{a}=1; tap {|#{a}|}")}
+  end
+
+  def test_unused_variable
+    o = Object.new
+    assert_warning(/assigned but unused variable/) {o.instance_eval("def foo; a=1; nil; end")}
+    a = "\u{3042}"
+    assert_warning(/#{a}/) {o.instance_eval("def foo; #{a}=1; nil; end")}
+  end
+
+  def test_named_capture_conflict
+    a = 1
+    assert_warning(/named capture conflict/) {eval("a = 1; /(?<a>)/ =~ ''")}
+    a = "\u{3042}"
+    assert_warning(/#{a}/) {eval("#{a} = 1; /(?<#{a}>)/ =~ ''")}
+  end
+
+=begin
+  def test_past_scope_variable
+    assert_warning(/past scope/) {catch {|tag| eval("BEGIN{throw tag}; tap {a = 1}; a")}}
+  end
+=end
 end
